/*
 * import.cpp
 *
 *  Created on: 29 Mar 2022
 *      Author: michi
 */

#include "../kaba.h"
#include "Parser.h"
#include "../../os/filesystem.h"
#include "../../os/msg.h"


#include "../../config.h"

#ifdef _X_USE_HUI_
#include "../../hui/Application.h"
#elif defined(_X_USE_HUI_MINIMAL_)
#include "../../hui_minimal/Application.h"
#endif


const int MAX_IMPORT_DIRECTORY_PARENTS = 5;

namespace kaba {

extern Array<shared<Module>> loading_module_stack;
void SetImmortal(SyntaxTree *ps);

string canonical_import_name(const string &s) {
	return s.lower().replace(" ", "").replace("_", "");
}

string dir_has(const Path &dir, const string &name) {
	auto list = os::fs::search(dir, "*", "fd");
	for (auto &e: list)
		if (canonical_import_name(e.str()) == name)
			return e.str();
	return "";
}

Path import_dir_match(const Path &dir0, const string &name) {
	auto xx = name.replace(".kaba", "").explode("/");
	Path filename = dir0;

	// parents matching?
	for (int i=0; i<xx.num-1; i++) {
		string e = dir_has(filename, canonical_import_name(xx[i]));
		if (e == "")
			return Path::EMPTY;
		filename |= e;
	}
	{
		// direct file  zzz.kaba?
		string e = dir_has(filename, canonical_import_name(xx.back() + ".kaba"));
		if (e != "") {
			filename |= e;
			return filename;
		}

		// package  zzz/zzz.kaba  or  zzz/__main__.kaba?
		e = dir_has(filename, canonical_import_name(xx.back()));
		if (e == "")
			return Path::EMPTY;
		filename |= e;
		if (os::fs::exists(filename | "__main__.kaba"))
			return filename | "__main__.kaba";
		if (os::fs::exists(filename | (xx.back() + ".kaba")))
			return filename | (xx.back() + ".kaba");
		if (os::fs::exists(filename | "main.kaba"))
			return filename | "main.kaba";
		return Path::EMPTY;
	}
	return filename;

	if (os::fs::exists(dir0 | name))
		return dir0 | name;
	return Path::EMPTY;
}

Path find_installed_lib_import(const string &name) {
	Path kaba_dir = hui::Application::directory.parent() | "kaba";
	if (hui::Application::directory.basename()[0] == '.')
		kaba_dir = hui::Application::directory.parent() | ".kaba";
	Path kaba_dir_static = hui::Application::directory_static.parent() | "kaba";
	for (auto &dir: Array<Path>({kaba_dir, kaba_dir_static})) {
		auto path = (dir | "lib" | name).canonical();
		if (os::fs::exists(path))
			return path;
	}
	return Path::EMPTY;
}

Path find_import(Module *s, const string &_name) {
	string name = _name.replace(".kaba", "");
	name = name.replace(".", "/") + ".kaba";

	// deprecated...
	if (name.head(2) == "@/")
		return find_installed_lib_import(name.sub(2));

	for (int i=0; i<MAX_IMPORT_DIRECTORY_PARENTS; i++) {
		Path filename = import_dir_match((s->filename.parent() | string("../").repeat(i)).canonical(), name);
		if (filename)
			return filename;
	}

	// installed?
	return find_installed_lib_import(name);

	return Path::EMPTY;
}

shared<Module> get_import(Parser *parser, const string &name, int token_id) {

	// internal packages?
	for (auto p: parser->context->packages)
		if (p->filename.str() == name)
			return p;

	Path filename = find_import(parser->tree->module, name);
	if (!filename)
		parser->do_error(format("can not find import '%s'", name), token_id);

	for (auto ss: weak(loading_module_stack))
		if (ss->filename == filename)
			parser->do_error("recursive import", token_id);

	msg_right();
	shared<Module> include;
	try {
		include = parser->context->load_module(filename, parser->tree->module->just_analyse or config.compile_os);
		// os-includes will be appended to syntax_tree... so don't compile yet
	} catch (kaba::Exception &e) {
		msg_left();

		auto p = new Exception(e);
		e.line = parser->Exp.token_physical_line_no(token_id);
		e.column = parser->Exp.token_line_offset(token_id);
		e.filename = parser->tree->module->filename;
		e.parent = p;
		throw e;
	}

	msg_left();
	return include;
}


static bool _class_contains(const Class *c, const string &name) {
	for (auto *cc: weak(c->classes))
		if (cc->name == name)
			return true;
	for (auto *f: weak(c->functions))
		if (f->name == name)
			return true;
	for (auto *cc: weak(c->constants))
		if (cc->name == name)
			return true;
	return false;
}

void namespace_import_contents(Class *parent, const Class *child) {
	for (auto *c: weak(child->classes))
		parent->classes.add(c);
	for (auto *f: weak(child->functions))
		parent->functions.add(f);
	for (auto *v: weak(child->static_variables))
		parent->static_variables.add(v);
	for (auto *c: weak(child->constants))
		parent->constants.add(c);
}

Class *get_namespace_for_import(SyntaxTree *tree, const string &name) {
	auto xx = name.explode(".");
	Class *ns = tree->imported_symbols.get();
	flags_set(ns->flags, Flags::EXTERN); // "don't delete contents..."

	auto get_next = [tree] (Class *ns, const string &name) {
		for (auto c: weak(ns->classes))
			if (c->name == name)
				return const_cast<Class*>(c);
		return tree->create_new_class(name, Class::Type::REGULAR, 0, 0, nullptr, {}, ns, -1);
	};

	for (auto &x: xx)
		ns = get_next(ns, x);
	return ns;
}

// import data from an included module file
//   indirect: "import" => true, "use" => false
void SyntaxTree::import_data(shared<Module> source, bool indirect, const string &as_name) {
	for (auto i: weak(includes))
		if (i == source)
			return;

	SyntaxTree *ps = source->tree.get();

	// propagate immortality TO the (dependent) source!
	//  (might be unnecessary due to shared pointers)
	if (flag_immortal)
		SetImmortal(ps);

	flag_string_const_as_cstring |= ps->flag_string_const_as_cstring;


	/*if (FlagCompileOS) {
		import_deep(this, ps);
	} else {*/
	if (indirect) {
		// "import"
		auto ns = get_namespace_for_import(this, as_name);
		namespace_import_contents(ns, ps->base_class);
	} else {
		// "use"
		namespace_import_contents(imported_symbols.get(), ps->base_class);
		if (source->is_system_module())
			if (!_class_contains(imported_symbols.get(), ps->base_class->name)) {
				imported_symbols->classes.add(ps->base_class);
			}

		// hack: package auto import
		for (auto c: weak(ps->base_class->constants))
			if (c->name == "EXPORT_IMPORTS") {
				for (auto i: weak(ps->includes))
					if (!i->is_system_module())
						import_data(i, indirect, "");
			}
	}
	includes.add(source);
	//}
}


}
